from builtins import str
from builtins import range
from builtins import object
import json
import os

import solnlib.utils as utils

from splunktaucclib.global_config import GlobalConfig, GlobalConfigSchema


'''
Usage Examples:
setup_util = Setup_Util(uri, session_key)
setup_util.get_log_level()
setup_util.get_proxy_settings()
setup_util.get_credential_account("my_account_name")
setup_util.get_customized_setting("my_customized_field_name")
'''

'''
setting object structure.
It is stored in self.__cached_global_settings
Note, this structure is only maintained in this util.
setup_util transforms global settings in os environment or from ucc into this structure.
{
    "proxy_settings": {
    "proxy_enabled": False/True,
    "proxy_url": "example.com",
    "proxy_port": "1234",
    "proxy_username": "",
    "proxy_password": "",
    "proxy_type": "http",
    "proxy_rdns": False/True
    },
    "log_settings": {
        "loglevel": "DEBUG"
    },
    "credential_settings": [{
        "name": "account_id",
        "username": "example_account",
        "password": "example_password"
    }, { # supported by ucc, not seen any usage in AoB
        "api_key": "admin",
        "api_uuid": "admin",
        "endpoint": "some url",
        "name": "account1"
    }],
    "customized_settings": {
        "text_name": "content",
        "pass_name": "password",
        "checkbox": 0/1
    }
}
'''

GLOBAL_SETTING_KEY = "global_settings"
AOB_TEST_FLAG = 'AOB_TEST'

PROXY_SETTINGS = "proxy_settings"
LOG_SETTINGS = "log_settings"
CREDENTIAL_SETTINGS = "credential_settings"
CUSTOMIZED_SETTINGS = "customized_settings"

UCC_PROXY = "proxy"
UCC_LOGGING = "logging"
UCC_CUSTOMIZED = "additional_parameters"
UCC_CREDENTIAL = "account"

CONFIGS = [CREDENTIAL_SETTINGS]
SETTINGS = [PROXY_SETTINGS, LOG_SETTINGS, CUSTOMIZED_SETTINGS]

PROXY_ENABLE_KEY = 'proxy_enabled'
PROXY_RDNS_KEY = 'proxy_rdns'
LOG_LEVEL_KEY = 'loglevel'
LOG_LEVEL_KEY_ENV = 'log_level'

TYPE_CHECKBOX = "checkbox"
ALL_SETTING_TYPES = ['text', 'password', 'checkbox', 'dropdownlist', 'multi_dropdownlist', 'radiogroup']


def get_schema_path():
    dirname = os.path.dirname
    basedir = dirname(dirname((dirname(__file__))))
    return os.path.join(basedir, 'appserver', 'static', 'js', 'build', 'globalConfig.json')


class Setup_Util(object):
    def __init__(self, uri, session_key, logger=None):
        self.__uri = uri
        self.__session_key = session_key
        self.__logger = logger
        self.scheme, self.host, self.port = utils.extract_http_scheme_host_port(
            self.__uri)
        self.__cached_global_settings = {}
        self.__global_config = None


    def init_global_config(self):
        if self.__global_config is not None:
            return
        schema_file = get_schema_path()
        if not os.path.isfile(schema_file):
            self.log_error("Global config JSON file not found!")
            self.__global_config = None
        else:
            with open(get_schema_path()) as f:
                json_schema = ''.join([l for l in f])
            self.__global_config = GlobalConfig(self.__uri, self.__session_key,
                                                GlobalConfigSchema(json.loads(json_schema)))

    def log_error(self, msg):
        if self.__logger:
            self.__logger.error(msg)

    def log_info(self, msg):
        if self.__logger:
            self.__logger.info(msg)

    def log_debug(self, msg):
        if self.__logger:
            self.__logger.debug(msg)

    def _parse_conf(self, key):
        if os.environ.get(AOB_TEST_FLAG, 'false') == 'true':
            global_settings = self._parse_conf_from_env(json.loads(os.environ.get(GLOBAL_SETTING_KEY, '{}')))
            return global_settings.get(key)
        else:
            return self._parse_conf_from_global_config(key)

    def _parse_conf_from_env(self, global_settings):
        '''
        this is run in test env
        '''
        if not self.__cached_global_settings:
            # format the settings, the setting from env is from global_setting
            # meta
            self.__cached_global_settings = {}
            for s_k, s_v in list(global_settings.items()):
                if s_k == PROXY_SETTINGS:
                    proxy_enabled = s_v.get(PROXY_ENABLE_KEY)
                    proxy_rdns = s_v.get(PROXY_RDNS_KEY)
                    if type(proxy_enabled) != bool:
                        s_v[PROXY_ENABLE_KEY] = utils.is_true(proxy_enabled)
                    if type(proxy_rdns) != bool:
                        s_v[PROXY_RDNS_KEY] = utils.is_true(proxy_rdns)
                    self.__cached_global_settings[PROXY_SETTINGS] = s_v
                elif s_k == LOG_SETTINGS:
                    self.__cached_global_settings[LOG_SETTINGS] = {
                        LOG_LEVEL_KEY: s_v.get(LOG_LEVEL_KEY_ENV)
                    }
                elif s_k == CREDENTIAL_SETTINGS:
                    # add account id to accounts
                    for i in range(0, len(s_v)):
                        s_v[i]['name'] = 'account' + str(i)
                    self.__cached_global_settings[CREDENTIAL_SETTINGS] = s_v
                else:  # should be customized settings
                    self.__cached_global_settings[CUSTOMIZED_SETTINGS] = {}
                    for s in s_v:
                        field_type = s.get('type')
                        if not field_type:
                            self.log_error(
                                'unknown type for customized var:{}'.format(s))
                            continue
                        self.__cached_global_settings['customized_settings'][s.get('name', '')] = self._transform(
                            s.get("value", ""), field_type)

        return self.__cached_global_settings

    def _parse_conf_from_global_config(self, key):
        if self.__cached_global_settings and key in self.__cached_global_settings:
            return self.__cached_global_settings.get(key)
        self.init_global_config()
        if self.__global_config is None:
            return None
        if key in CONFIGS:
            accounts = self.__global_config.configs.load().get(UCC_CREDENTIAL, [])
            if accounts:
                for account in accounts:
                    if 'disabled' in account:
                        del account['disabled']
            self.__cached_global_settings[CREDENTIAL_SETTINGS] = accounts
        elif key in SETTINGS:
            settings = self.__global_config.settings.load()
            self.__cached_global_settings.update({UCC_PROXY: None, UCC_LOGGING: None, UCC_CUSTOMIZED: None})
            customized_setting = {}
            for setting in settings.get('settings', []):
                # filter out disabled setting page and 'disabled' field
                if setting.get('disabled', False):
                    continue
                if setting['name'] == UCC_LOGGING:
                    self.__cached_global_settings[LOG_SETTINGS] = {
                        LOG_LEVEL_KEY: setting.get(LOG_LEVEL_KEY)
                    }
                elif setting['name'] == UCC_PROXY:
                    if 'disabled' in setting:
                        del setting['disabled']
                    setting[PROXY_ENABLE_KEY] = utils.is_true(setting.get(PROXY_ENABLE_KEY, '0'))
                    setting[PROXY_RDNS_KEY] = utils.is_true(setting.get(PROXY_RDNS_KEY, '0'))
                    self.__cached_global_settings[PROXY_SETTINGS] = setting
                else:  # should be customized settings
                    if 'disabled' in setting:
                        del setting['disabled']
                    customized_setting.update(setting)
            self.__cached_global_settings[CUSTOMIZED_SETTINGS] = customized_setting

        return self.__cached_global_settings.get(key)

    def get_log_level(self):
        log_level = "INFO"
        log_settings = self._parse_conf(LOG_SETTINGS)
        if log_settings is None:
            self.log_info("Log level is not set, use default INFO")
        else:
            log_level = log_settings.get(LOG_LEVEL_KEY, None)
            if not log_level:
                self.log_info("Log level is not set, use default INFO")
                log_level = "INFO"
        return log_level

    def get_proxy_settings(self):
        proxy_settings = self._parse_conf(PROXY_SETTINGS)
        if proxy_settings is None:
            self.log_info("Proxy is not set!")
            return {}
        proxy_enabled = proxy_settings.get(PROXY_ENABLE_KEY)
        if not proxy_enabled:
            self.log_info("Proxy is not enabled!")
            return {}
        proxy_settings = {
            "proxy_url": proxy_settings.get("proxy_url", ""),
            "proxy_port": proxy_settings.get("proxy_port", None),
            "proxy_username": proxy_settings.get("proxy_username", ""),
            "proxy_password": proxy_settings.get("proxy_password", ""),
            "proxy_type": proxy_settings.get("proxy_type", ""),
            "proxy_rdns": proxy_settings.get("proxy_rdns")
        }
        self._validate_proxy_settings(proxy_settings)
        return proxy_settings

    def get_credential_by_id(self, account_id):
        credential_settings = self._parse_conf(CREDENTIAL_SETTINGS)
        for account in credential_settings:
            if account.get('name', None) == account_id:
                return account
        self.log_error("Credential account with account id {} can not be found".format(account_id))
        return None

    def get_credential_by_username(self, username):
        credential_settings = self._parse_conf(CREDENTIAL_SETTINGS)
        for account in credential_settings:
            if account.get('username', None) == username:
                return account
        self.log_error("Credential account with username {} can not be found".format(username))
        return None

    def get_customized_setting(self, key):
        customized_settings = self._parse_conf(CUSTOMIZED_SETTINGS)
        if customized_settings is None:
            self.log_info("Customized setting is not set")
            return None
        if key not in customized_settings:
            self.log_info("Customized key can not be found")
            return None
        customized_setting = customized_settings.get(key, None)
        if customized_setting is None:
            self.log_error("Cannot find customized setting with key %s" % key)
        return customized_setting

    def _validate_proxy_settings(self, proxy_settings):
        if proxy_settings:
            if proxy_settings.get('proxy_url') == "":
                raise Exception("Proxy host must not be empty!")
            proxy_port = proxy_settings.get('proxy_port')
            if proxy_port is None or not proxy_port.isdigit():
                raise Exception("Proxy port must be a number!")

    def _transform(self, value, field_type):
        '''
        This is method is only used when parsing customized global params from env.
        Only checkbox type needs transform. Other types will be extracted automatically when apply json.loads.
        :param value:
        :param field_type: can be checkbox, text, password, dropdownlist, multi_dropdownlist, radiogroup
        :return:
        '''
        if field_type == TYPE_CHECKBOX:
            return utils.is_true(value)
        elif field_type in ALL_SETTING_TYPES:
            return value
        else:
            raise Exception("Type of this customized setting is corrupted. Value: {}, type: {}"
                            .format(value, field_type))


    '''
    # the following methods is used by AoB internally
    # user should not use this
    # These methods returns the similiar structure like ucc libs

    the output of config is like
{
  "account": [
    {
      "username": "admin",
      "credential": "a",
      "name": "ddddd",
      "disabled": false
    }
  ]
}

    the output of settings is like
{
  "settings": [
    {
      "additional_parameters": {
        "checkbox": "1",
        "text": "msn",
        "disabled": false
      }
    },
    {
      "proxy": {
        "proxy_type": "http",
        "proxy_port": "9999",
        "proxy_url": "localhost",
        "proxy_rdns": "1",
        "disabled": false,
        "proxy_password": "a",
        "proxy_username": "admin",
        "proxy_enabled": "1"
      }
    },
    {
      "logging": {
        "loglevel": "ERROR",
        "disabled": false
      }
    }
  ]
}
    '''
    def get_ucc_log_setting(self):
        return {UCC_LOGGING: self._parse_conf(LOG_SETTINGS)}

    def get_ucc_proxy_setting(self):
        p = dict(self.get_proxy_settings())
        p[PROXY_ENABLE_KEY] = True if p else False
        return {
            UCC_PROXY: p
        }


    def get_ucc_customized_setting(self):
        customized_settings = self._parse_conf(CUSTOMIZED_SETTINGS)
        if customized_settings:
            return {
                UCC_CUSTOMIZED: customized_settings
            }
        else:
            return {}

    # account belongs to the configs
    def get_ucc_account_config(self):
        return {
            UCC_CREDENTIAL: self._parse_conf(CREDENTIAL_SETTINGS)
        }



__version__ = "4.0.4"

import logging
import os
import glob
from os import system
import shutil
import argparse
import json
from .uccrestbuilder.global_config import (
    GlobalConfigBuilderSchema,
    GlobalConfigPostProcessor,
)
from .uccrestbuilder import build
from .start_alert_build import alert_build

from jinja2 import Environment, FileSystemLoader

outputdir = os.path.join(os.getcwd(), "output")
sourcedir = os.path.dirname(os.path.realpath(__file__))

j2_env = Environment(
    loader=FileSystemLoader(os.path.join(sourcedir, "templates"))
)

logger = logging.getLogger('UCC')
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s [%(name)s] %(levelname)s: %(message)s')
shandler = logging.StreamHandler()
shandler.setLevel(logging.INFO)
shandler.setFormatter(formatter)
logger.addHandler(shandler)

PARENT_DIR = ".."


def get_os_path(path):
    if "\\\\" in path:
        path = path.replace("\\\\", os.sep)
    else:
        path = path.replace("\\", os.sep)
    path = path.replace("/", os.sep)
    return path.strip(os.sep)

def recursive_overwrite(src, dest, ignore_list=None):
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        for f in files:
            if not ignore_list or not os.path.join(dest, f) in ignore_list:
                recursive_overwrite(
                    os.path.join(src, f), os.path.join(dest, f), ignore_list
                )
            else:
                logger.info("Excluding : {}".format(os.path.join(dest, f)))
    else:
        if os.path.exists(dest):
            os.remove(dest)
        shutil.copy(src, dest)


def clean_before_build():

    logger.info("Cleaning out directory " + outputdir)
    shutil.rmtree(os.path.join(outputdir), ignore_errors=True)
    os.makedirs(os.path.join(outputdir))
    logger.info("Cleaned out directory " + outputdir)


def copy_package_source(args, ta_name):
    logger.info("Copy package directory ")
    recursive_overwrite(args.source, os.path.join(outputdir, ta_name))


def export_package(args, ta_name, ignore_list=None):
    logger.info("Exporting package")
    recursive_overwrite(os.path.join(outputdir, ta_name), args.source, ignore_list)
    logger.info("Final build ready at: {}".format(outputdir))


def copy_package_template(args, ta_name):
    logger.info("Copy UCC template directory")
    recursive_overwrite(
        os.path.join(sourcedir,"package"), os.path.join(outputdir, ta_name)
    )


def replace_token(args, ta_name):
    # replace token in template
    logger.info("Replace tokens in views")
    views = ["inputs.xml", "configuration.xml", "redirect.xml"]
    for view in views:
        template_dir = os.path.join(
            outputdir, ta_name, "default", "data", "ui", "views"
        )
        with open(os.path.join(template_dir, view)) as f:
            s = f.read()

        # Safely write the changed content, if found in the file
        with open(os.path.join(template_dir, view), "w") as f:
            s = s.replace("${package.name}", ta_name)
            if view == "redirect.xml":
                s = s.replace("${ta.name}", ta_name.lower())
            f.write(s)

def install_libs(parent_path, ucc_lib_target):

    def _install_libs(requirements, ucc_target, installer="python3"):
        
        if not os.path.exists(requirements):
            logging.warning("Unable to find requirements file. {}".format(requirements))
        else:
            if not os.path.exists(ucc_target):
                os.makedirs(ucc_target)
            install_cmd = (
                installer +" -m pip install -r \""
                + requirements
                + "\" --no-compile --no-binary :all: --target \""
                + ucc_target
                + "\""
            )
            os.system(install_cmd)
            remove_files(ucc_target)

    if os.path.exists(os.path.join(parent_path, "requirements.txt")):
        _install_libs(requirements=os.path.join(parent_path, "requirements.txt"), ucc_target=ucc_lib_target)

    if os.path.exists(os.path.join(parent_path, "requirements_py2.txt")):
        _install_libs(requirements=os.path.join(parent_path, "requirements_py2.txt"), installer="python2", ucc_target=os.path.join(ucc_lib_target, "py2"))

    if os.path.exists(os.path.join(parent_path, "requirements_py3.txt")):
        _install_libs(requirements=os.path.join(parent_path, "requirements_py3.txt"), ucc_target=os.path.join(ucc_lib_target, "py3"))


def remove_files(path):
    rmdirs = glob.glob(os.path.join(path, "*.egg-info")) + glob.glob(os.path.join(path, "*.dist-info"))
    for rmdir in rmdirs:
        shutil.rmtree(rmdir)

def copy_splunktaucclib(args, ta_name):
    logger.info("Copy splunktaucclib directory")
    recursive_overwrite(
        os.path.join(sourcedir, "splunktaucclib"),
        os.path.join(outputdir, ta_name, "lib", "splunktaucclib"),
    )


def generate_rest(args, ta_name, scheme, import_declare_name):
    build(
        scheme,
        "splunktaucclib.rest_handler.admin_external.AdminExternalHandler",
        os.path.join(outputdir, ta_name),
        j2_env,
        post_process=GlobalConfigPostProcessor(),
        import_declare_name=import_declare_name,
    )


def is_oauth_configured(ta_tabs):
    # check if oauth is configured in globalConfig.json
    for tab in ta_tabs:
        if tab["name"] == "account":
            for elements in tab["entity"]:
                if elements["type"] == "oauth":
                    return True
            break
    return False


def replace_oauth_html_template_token(args, ta_name, ta_version):
    html_template_path = os.path.join(
        outputdir, ta_name, "appserver", "templates"
    )
    with open(os.path.join(html_template_path, "redirect.html")) as f:
        s = f.read()

    # Safely write the changed content, if found in the file
    with open(os.path.join(html_template_path, "redirect.html"), "w") as f:
        # replace addon name in html template
        s = s.replace("${ta.name}", ta_name.lower())
        # replace addon version in html template
        s = s.replace("${ta.version}", ta_version)
        f.write(s)


def modify_and_replace_token_for_oauth_templates(
    args, ta_name, ta_tabs, ta_version
):
    redirect_xml_src = os.path.join(
        outputdir, ta_name, "default", "data", "ui", "views", "redirect.xml"
    )
    redirect_js_src = os.path.join(
        outputdir, ta_name, "appserver", "static", "js", "build", "redirect_page.js"
    )
    redirect_html_src = os.path.join(
        outputdir, ta_name, "appserver", "templates", "redirect.html"
    )
    # if oauth is configured replace token in html template and rename the templates with respect to addon name
    if is_oauth_configured(ta_tabs):
        replace_oauth_html_template_token(args, ta_name, ta_version)

        redirect_js_dest = (
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "")
            + ta_name.lower()
            + "_redirect_page."
            + ta_version
            + ".js"
        )
        redirect_html_dest = (
            os.path.join(outputdir, ta_name, "appserver", "templates", ta_name.lower() + "_redirect.html")
        )
        redirect_xml_dest = (
            os.path.join(outputdir, ta_name, "default", "data", "ui", "views", ta_name.lower() + "_redirect.xml")   
        )
        os.rename(redirect_js_src, redirect_js_dest)
        os.rename(redirect_html_src, redirect_html_dest)
        os.rename(redirect_xml_src, redirect_xml_dest)

     # if oauth is not configured remove the extra template
    else:
        os.remove(redirect_xml_src)
        os.remove(redirect_html_src)
        os.remove(redirect_js_src)

def add_modular_input(
    args, ta_name, schema_content, import_declare_name, j2_env
):

    services = schema_content.get("pages").get("inputs").get("services")
    for service in services:
        input_name = service.get("name")
        class_name = input_name.upper()
        description = service.get("title")
        entity = service.get("entity")
        field_white_list = ["name", "index", "sourcetype"]
        # filter fields in white list
        entity = [x for x in entity if x.get("field") not in field_white_list]
        import_declare = "import " + import_declare_name

        content = j2_env.get_template("input.template").render(
            import_declare=import_declare,
            input_name=input_name,
            class_name=class_name,
            description=description,
            entity=entity,
        )
        input_file_name = os.path.join(
            outputdir, ta_name, "bin", input_name + ".py"
        )
        with open(input_file_name, "w") as input_file:
            input_file.write(content)


def make_modular_alerts(args, ta_name, ta_namespace, schema_content):
    if schema_content.get("alerts"):

        alert_build(
            {"alerts": schema_content["alerts"]},
            ta_name,
            ta_namespace,
            outputdir,
            sourcedir,
        )
        
def get_ignore_list(args, path):
    if not os.path.exists(path):
        return []
    else:
        with open(path) as ignore_file:
            ignore_list = ignore_file.readlines()
        ignore_list = [(os.path.join(args.source, get_os_path(path))).strip() for path in ignore_list]
        return ignore_list


def main():
    parser = argparse.ArgumentParser(description="Build the add-on")
    parser.add_argument(
        "--source",
        type=str,
        nargs='?',
        help="Folder containing the app.manifest and app source",
        default="package",
    )
    parser.add_argument(
        "--config",
        type=str,
        nargs='?',
        help="Path to configuration file, Defaults to GlobalConfig.json in parent directory of source provided",
        default=None
    )
    args = parser.parse_args()
    
    if not os.path.exists(args.source):
        raise NotADirectoryError("{} not Found.".format(os.path.abspath(args.source)))

    # Setting default value to Config argument
    if not args.config:
        args.config = os.path.abspath(os.path.join(args.source, PARENT_DIR, "globalConfig.json"))

    clean_before_build()
    
    ignore_list = get_ignore_list(args, os.path.abspath(os.path.join(args.source, PARENT_DIR, ".uccignore")))
    if os.path.exists(args.config):

        with open(args.config, "r") as config_file:
            schema_content = json.load(config_file)

        scheme = GlobalConfigBuilderSchema(schema_content, j2_env)
        ta_name = schema_content.get("meta").get("name")
        ta_version = schema_content.get("meta").get("version")
        ta_tabs = schema_content.get("pages").get("configuration").get("tabs")
        ta_namespace = schema_content.get("meta").get("restRoot")
        import_declare_name = "import_declare_test"

        logger.info("Package ID is " + ta_name)

        copy_package_template(args, ta_name)


        shutil.copyfile(
            args.config,
            os.path.join(outputdir, ta_name, "appserver", "static", "js", "build", "globalConfig.json"),
        )
        ucc_lib_target = os.path.join(outputdir, ta_name, "lib")

        install_libs(
            parent_path=os.path.abspath(os.path.join(args.source, PARENT_DIR)),
            ucc_lib_target=ucc_lib_target
        )

        install_libs(
            parent_path=sourcedir,
            ucc_lib_target=ucc_lib_target
        )
        copy_splunktaucclib(args, ta_name)


        replace_token(args, ta_name)

        generate_rest(args, ta_name, scheme, import_declare_name)

        modify_and_replace_token_for_oauth_templates(
                args, ta_name, ta_tabs, schema_content.get('meta').get('version')
            )

        add_modular_input(
                args, ta_name, schema_content, import_declare_name, j2_env
            )

        make_modular_alerts(args, ta_name, ta_namespace, schema_content)

    else:
        logger.warning("Skipped installing UCC required python modules as GlobalConfig.json does not exist.")
        logger.warning("Skipped Generating UI components as GlobalConfig.json does not exist.")
        logger.info("Setting TA name as generic")
        
        ta_name = "TA-generic"
        ucc_lib_target = os.path.join(outputdir, ta_name, "lib")

        install_libs(
            parent_path=os.path.abspath(os.path.join(args.source, PARENT_DIR)),
            ucc_lib_target=ucc_lib_target
        )

    copy_package_source(args, ta_name)
    export_package(args, ta_name, ignore_list)


# encoding = utf-8

import logging
import sys

def get_stderr_stream_logger(logger_name=None, log_level=logging.INFO):
    if logger_name is None:
        logger_name = 'aob_default_logger'
    logger = logging.getLogger(logger_name)
    formatter = logging.Formatter('%(asctime)s - %(name)s - [%(levelname)s] - %(message)s')
    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    stderr_handler.setLevel(logging.DEBUG)
    stderr_handler.setFormatter(formatter)
    logger.addHandler(stderr_handler)
    logger.setLevel(log_level)
    return logger



from __future__ import absolute_import
# encoding = utf-8

from past.builtins import basestring
from builtins import str
from six import string_types as basestring
import functools

from . import monitor
import time

__all__ = ['initialize_metric_collector', 'function_run_time']


def initialize_metric_collector(config, update_config=False):
    '''
    config is a configuration dict.
     - app: required field, define the app name for the monitor
     - event_writer: event writer type, must be one of ['file', 'hec']
     - writer_config: a dict, define the configuration for the event writer.
                     different event writer has different configurations.
    update_config: force to update the configure
    '''
    m = monitor.Monitor().configure(config, force_update=update_config)
    m.start()


def write_event(ev, tags=[]):
    monitor.Monitor().write_event(ev, tags)


CREDENTIAL_KEYS = ['password', 'passwords', 'token']


def mask_credentials(data):
    '''
    The argument will be cloned
    '''
    masked_str = '******'
    if isinstance(data, dict):
        new_data = {}
        for k in list(data.keys()):
            if isinstance(k, str):
                _key = k.lower()
                sensitive_word = False
                for w in CREDENTIAL_KEYS:
                    if w in _key:
                        sensitive_word = True
                        break
                if sensitive_word:
                    new_data[k] = masked_str
                else:
                    new_data[k] = mask_credentials(data[k])
            else:
                new_data[k] = mask_credentials(data[k])
        return new_data
    elif isinstance(data, (list, tuple)):
        new_data = []
        for d in data:
            new_data.append(mask_credentials(d))
        return new_data
    elif isinstance(data, (str, basestring)):
        d = data.lower()
        sensitive_word = False
        for w in CREDENTIAL_KEYS:
            if w in d:
                sensitive_word = True
                break
        if sensitive_word:
            return masked_str
    elif not isinstance(data, (int, float)):
        return 'Class:' + data.__class__.__name__
    return data


max_length = 2048


def function_run_time(tags=[]):
    def apm_decorator(func):
        @functools.wraps(func)
        def func_wrappers(*args, **kwargs):
            m = monitor.Monitor()
            func_attr = {'function_name': func.__name__,
                         'positional_args': str(mask_credentials(args))[0:max_length],
                         'keyword_arguments': str(mask_credentials(kwargs))[0:max_length]}
            ev = {'action': 'invoke'}
            ev.update(func_attr)
            m.write_event(ev, tags)
            before_invoke = int(time.time() * 1000)
            ret = func(*args, **kwargs)
            after_invoke = int(time.time() * 1000)
            ev = {'action': 'done',
                  'time_cost': (after_invoke - before_invoke)}
            ev.update(func_attr)
            m.write_event(ev, tags)
            return ret

        return func_wrappers

    return apm_decorator


# TODO: should put this into unit test
# class A(object):
#     def __init__(self):
#         self.password = 'abc'
#
# if __name__ == "__main__":
#     aa = {'a': {'token': '123'}, 'b': A()}
#     print mask_credentials(aa)
#     print aa
#     print mask_credentials(123)
#     bb = {'l': [{'token': 123, 'password': '444'}]}
#     print mask_credentials(bb)
#     print bb



from __future__ import absolute_import


from builtins import object
class RestmapConf(object):

    _admin_template = """
[admin:{namespace}]
match = /{admin_match}
members = {endpoints}
"""

    _external_template = """
[admin_external:{name}]
handlertype = python
handlerfile = {rh_name}.py
handleractions = {actions}
handlerpersistentmode = true
"""

    @classmethod
    def build(cls, endpoints, namespace, admin_match):
        if not endpoints:
            return ''
        externals = [
            cls._admin_template.format(
                namespace=namespace,
                admin_match=admin_match,
                endpoints=', '.join([ep.name for ep in endpoints])
            )
        ]
        for endpoint in endpoints:
            external = cls._external_template.format(
                name=endpoint.name,
                rh_name=endpoint.rh_name,
                actions=', '.join(endpoint.actions()),
            )
            externals.append(external)
        return ''.join(externals)

    @classmethod
    def admin_externals(cls, endpoints):
        return [endpoint.name for endpoint in endpoints]


class WebConf(object):

    _template = """
[expose:{name}]
pattern = {name}
methods = POST, GET
"""

    _specified_template = """
[expose:{name}_specified]
pattern = {name}/*
methods = POST, GET, DELETE
"""

    _internal_template = """
[expose:{name}]
pattern = {endpoint}
methods = GET
"""

    @classmethod
    def build(cls, endpoints):
        stanzas = []
        for endpoint in endpoints:
            stanzas.append(
                cls._template.format(
                    namespace=endpoint.namespace,
                    name=endpoint.name,
                )
            )
            stanzas.append(
                cls._specified_template.format(
                    namespace=endpoint.namespace,
                    name=endpoint.name,
                )
            )
        # add splunkd data endpoint
        stanzas.append(
            cls._internal_template.format(
                name='_splunkd_data',
                endpoint='data/*'
            )
        )
        return ''.join(stanzas)



"""
REST Schema
"""

from __future__ import absolute_import

from builtins import object
from abc import abstractproperty

__all__ = [
    'RestSchemaError',
    'RestSchema',
]


class RestSchemaError(Exception):
    pass


class RestSchema(object):
    """
    REST Scheme.
    """

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def endpoint_name(name, namespace):
        return '{}_{}'.format(namespace, name)

    @abstractproperty
    def product(self):
        pass

    @abstractproperty
    def namespace(self):
        pass

    @abstractproperty
    def version(self):
        pass



from __future__ import absolute_import

from .base import RestEndpointBuilder


"""
This class is used to generate the endpoint for getting access token from auth code and 
extends RestEndpointBuilder class
"""


class OAuthModelEndpointBuilder(RestEndpointBuilder):

    """
    This is initialization of this endpoint builder class
    """

    def __init__(self, name, j2_env, namespace, **kwargs):
        super(OAuthModelEndpointBuilder, self).__init__(name, namespace, **kwargs)
        self._app_name = kwargs.get("app_name")
        self.j2_env = j2_env

    """
    Action will return the possible action for the endpoint
    """

    def actions(self):
        return ["edit"]

    """
    This will actually populate the jinja template with the token values and return it
    """

    def generate_rh(self, handler):
        return self.j2_env.get_template("oauth.template").render(
            app_name=self._app_name,
        )



"""
Error Handling.
"""

from __future__ import absolute_import

__all__ = ['STATUS_CODES', 'RestError']


# HTTP status codes
STATUS_CODES = {
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    411: 'Length Required',
    500: 'Internal Server Error',
    503: 'Service Unavailable'
}


class RestError(Exception):
    """
    REST Error.
    """

    def __init__(self, status, message):
        self.status = status
        self.reason = STATUS_CODES.get(
            status,
            'Unknown Error',
        )
        self.message = message
        err_msg = 'REST Error [%(status)s]: %(reason)s -- %(message)s' % {
            'status': self.status,
            'reason': self.reason,
            'message': self.message
        }
        super(RestError, self).__init__(err_msg)




from __future__ import absolute_import

import traceback

from ..rest_handler.schema import RestSchema, RestSchemaError


class GlobalConfigSchema(RestSchema):

    def __init__(self, content, *args, **kwargs):
        """

        :param content: Python object for Global Config Schema
        :param args:
        :param kwargs:
        """
        super(GlobalConfigSchema, self).__init__(*args, **kwargs)
        self._content = content
        self._inputs = []
        self._configs = []
        self._settings = []

        try:
            self._parse()
        except Exception:
            raise RestSchemaError(
                'Invalid Global Config Schema: %s' % traceback.format_exc(),
            )

    @property
    def product(self):
        return self._meta['name']

    @property
    def namespace(self):
        return self._meta['restRoot']

    @property
    def admin_match(self):
        return ''

    @property
    def version(self):
        return self._meta['apiVersion']

    @property
    def inputs(self):
        return self._inputs

    @property
    def configs(self):
        return self._configs

    @property
    def settings(self):
        return self._settings

    def _parse(self):
        self._meta = self._content['meta']
        pages = self._content['pages']
        self._parse_configuration(pages.get('configuration'))
        self._parse_inputs(pages.get('inputs'))

    def _parse_configuration(self, configurations):
        if not configurations or 'tabs' not in configurations:
            return
        for configuration in configurations['tabs']:
            if 'table' in configuration:
                self._configs.append(configuration)
            else:
                self._settings.append(configuration)

    def _parse_inputs(self, inputs):
        if not inputs or 'services' not in inputs:
            return
        self._inputs = inputs['services']
